open import Agda.Primitive renaming (Set to Type)

open import Relation.Binary.PropositionalEquality hiding ([_])

open import Data.Bool
open import Data.Nat

import Data.List as L
import Data.Vec as V
import Data.String as S
import Data.Char as C
open L using (List)
open V using (Vec)
open S using (String)
open C using (Char)

open import Base
open import Bits-and-Bytes
open import NonEmpty

module ASCII where

list-is-ascii : List Char → Bool
list-is-ascii L.[] = true
list-is-ascii (c L.∷ chs) = (C.toℕ c <ᵇ 128) ∧ list-is-ascii chs

string-is-ascii : String → Type
string-is-ascii x = T (list-is-ascii (S.toList x))

encode-vec : (s : String) → {string-is-ascii s} → Vec Byte (S.length s)
encode-vec x {prf} = go (S.toList x)
  where
    go : (l : List Char) → Vec Byte (L.length l)
    go L.[] = V.[]
    go (ch L.∷ chs) = mkClip (C.toℕ ch) V.∷ go chs

encode-list⁺ : (s : String) → {string-is-ascii s} → {T (0 <ᵇ S.length s)} → List⁺ Byte
encode-list⁺ s {ascii} {nonempty} = vec-to-list⁺ v
  where
    open import Data.Nat.Properties
    open S

    len≢0 : length s ≢ 0
    len≢0 = m<n⇒n≢0 (<ᵇ⇒< 0 (length s) nonempty)

    v : Vec Byte (suc (pred (length s)))
    v = subst (λ x → Vec Byte x) (sym (suc[pred[n]]≡n {length s} len≢0)) (encode-vec s {ascii})

encode-list : (s : String) → {string-is-ascii s} → List Byte
encode-list s {ascii} = V.toList (encode-vec s {ascii})
