open import Agda.Primitive renaming (Set to Type)

open import Data.Nat
open import Data.Nat.DivMod
open import Data.Nat.Properties using (<ᵇ⇒<; *-comm; *-monoˡ-≤; *-monoˡ-<)
open import Data.Fin using () renaming (zero to zeroᶠ; suc to sucᶠ)
open import Data.Bool hiding (_<_; _≤_; _<?_)
open import Data.Vec
open import Data.Product
open import Relation.Binary.PropositionalEquality
open import Relation.Nullary

open import Base

module Bits-and-Bytes where

record Byte : Type where
  field
    value : ℕ
    {in-range} : value < 256
open Byte public

<_> : (v : ℕ) → {T (v <ᵇ 256)} → Byte
< v > {prf} = record { value = v ; in-range = <ᵇ⇒< v 256 prf }

module bytes where
  NULL CR LF SPACE BIG : Byte
  NULL    = < 0 >
  LF      = < 10 >
  CR      = < 13 >
  SPACE   = < 32 >
  SLASH   = < 47 >
  PERCENT = < 37 >
  BIG     = < 0xff >

private
  import Data.String as String
  import Data.Fin as Fin
  import Data.Char as Char

  toHexDigit : ℕ → String.String
  toHexDigit 0 = "0"
  toHexDigit 1 = "1"
  toHexDigit 2 = "2"
  toHexDigit 3 = "3"
  toHexDigit 4 = "4"
  toHexDigit 5 = "5"
  toHexDigit 6 = "6"
  toHexDigit 7 = "7"
  toHexDigit 8 = "8"
  toHexDigit 9 = "9"
  toHexDigit 10 = "a"
  toHexDigit 11 = "b"
  toHexDigit 12 = "c"
  toHexDigit 13 = "d"
  toHexDigit 14 = "e"
  toHexDigit 15 = "f"
  toHexDigit x = "?"

  toHex : ℕ → String.String
  toHex n = toHexDigit (DivMod.quotient divmod) String.++ toHexDigit (Fin.toℕ (DivMod.remainder divmod))
    where
      divmod : DivMod n 16
      divmod = n divMod 16


instance
  EqByte : Eq Byte
  EqByte ._==_ a b = a .value == b .value

  ShowByte : Show Byte
  ShowByte .show x =
    if (0x20 ≤ᵇ x .value) ∧ (x .value <ᵇ 0x7f) ∧ not (x .value == 0x60)
      then "`" String.++ String.fromChar (Char.fromℕ (x .value)) String.++ "`"
      else "<" String.++ toHex (x .value) String.++ ">"
    where
      import Data.String as String

  import Data.List as List

  ShowByteList : Show (List.List Byte)
  ShowByteList .show xs = "'" String.++ show-inner xs String.++ "'"
    where
      show-inner : List.List Byte → String.String
      show-inner List.[] = ""
      show-inner (x List.∷ xs) =
        if (0x20 ≤ᵇ x .value) ∧ (x .value <ᵇ 0x7f)
          then String.fromChar (Char.fromℕ (x .value)) String.++ show-inner xs
          else "\\x" String.++ toHex (x .value) String.++ show-inner xs

mkClip : ℕ → Byte
mkClip v = case v <? 256 of λ where
  (yes v<256) → record { value = v ; in-range = v<256 }
  (no _) → bytes.BIG

data Bit : Type where
  𝟘 𝟙 : Bit

-- LSB first‼
Bits : ℕ → Type
Bits n = Vec Bit n

Bit8 : Type
Bit8 = Bits 8

rshift1 : {max : ℕ} → (n : ℕ) → n < 2 * max → Σ (ℕ × Bit) λ (quot , _) → quot < max
rshift1 {max} n n<2m = (quot , shift-out) , quot<max
  where
    divmod : DivMod n 2
    divmod = n divMod 2

    quot : ℕ
    quot = DivMod.quotient divmod

    shift-out : Bit
    shift-out = case (DivMod.remainder divmod) of λ where
      (zeroᶠ) → 𝟘
      (sucᶠ zeroᶠ) → 𝟙

    postulate
      quot<max : quot < max
    -- quot<max = {!!} -- TODO

enbits : {n : ℕ} → (x : ℕ) → x < 2 ^ n → Bits n
enbits {zero} zero x<2^n = []
enbits {zero} (suc x) (s≤s ())
enbits {n@(suc pn)} x x<2^n = shift-out ∷ enbits {pn} quot quot<2^pn
  where
    a : (Σ (ℕ × Bit) λ (quot , _) → quot < 2 ^ pn)
    a = rshift1 {2 ^ pn} x x<2^n

    quot : ℕ
    quot = proj₁ (proj₁ a)

    shift-out : Bit
    shift-out = proj₂ (proj₁ a)

    quot<2^pn : quot < 2 ^ pn
    quot<2^pn = proj₂ a

unbits : {n : ℕ} → Bits n → Σ ℕ λ x → x < 2 ^ n
unbits {zero} [] = (0 , s≤s z≤n)
unbits {n@(suc pn)} (x ∷ xs) = crazy-hamburger
  where
    -- suc x ≤ n → suc (suc (2 * x)) ≤ 2 * n
    x<n→s2x<2n : {x n : ℕ} → x < n → suc (2 * x) < 2 * n
    x<n→s2x<2n {x} {n} x<n = subst₂ (λ a b → suc (suc a) ≤ b) (*-comm x 2) (*-comm n 2) (*-monoˡ-≤ 2 x<n)

    x<n→2x<2n : {x n : ℕ} → x < n → 2 * x < 2 * n -- 2 * x < 2 * n
    x<n→2x<2n {x} {n} x<n = subst₂ (λ a b → a < b) (*-comm x 2) (*-comm n 2) (*-monoˡ-< 1 x<n)

    rest : Σ ℕ λ x → x < 2 ^ pn
    rest = unbits {pn} xs

    crazy-hamburger : Σ ℕ λ x → x < 2 ^ n
    crazy-hamburger = case x of λ where
      𝟘 → (2 * proj₁ rest , x<n→2x<2n (proj₂ rest))
      𝟙 → (suc (2 * proj₁ rest) , x<n→s2x<2n (proj₂ rest))

Byte-to-Bit8 : Byte → Bit8
Byte-to-Bit8 b = enbits (b .value) (b .in-range)

Bit8-to-Byte : Bit8 → Byte
Bit8-to-Byte x =
  let b , b<256 = unbits x
  in record { value = b ; in-range = b<256 }

module Foreign where
  {-# FOREIGN GHC
    import qualified Data.Word as Word
  #-}

  postulate
    Word8 : Type
    ℕ-to-Word8 : ℕ → Word8 -- wraps
    Word8-to-ℕ : Word8 → ℕ

    Word8-to-ℕ-in-range : (w : Word8) → Word8-to-ℕ w < 256

  {-# COMPILE GHC Word8 = type Word.Word8 #-}
  {-# COMPILE GHC ℕ-to-Word8 = fromIntegral :: Integer -> Word.Word8 #-}
  {-# COMPILE GHC Word8-to-ℕ = fromIntegral :: Word.Word8 -> Integer #-}

  Byte-to-Word8 : Byte → Word8
  Byte-to-Word8 b = ℕ-to-Word8 (b .value)

  Word8-to-Byte : Word8 → Byte
  Word8-to-Byte w = record { value = Word8-to-ℕ w ; in-range = Word8-to-ℕ-in-range w }
open Foreign using (Word8; Byte-to-Word8; Word8-to-Byte) public
