open import Agda.Primitive renaming (Set to Type)

open import Data.Bool

import Data.String as S
import Data.List as L
import Data.Char as C
open S using (String)
open L using (List)
open C using (Char)

open import HTML
open import Base

module HTML.Render where

private
  -- encodes using HTML decimal escapes (&#621;)
  encode : List Char → String → String
  encode escapes text = S.fromList (L.concatMap (λ x → S.toList (escape-char x)) (S.toList text))
    where
      escape-char : Char → String
      escape-char ch =
        if L.any (_== ch) escapes
          then "&#" S.++ show (C.toℕ ch) S.++ ";"
          else S.fromChar ch

render-attr : Attribute → String
render-attr x = x .name S.++ "=" S.++ encode (S.toList "\t\n\r \"'=<>`") (x .value)
  where
    open Attribute

{-# TERMINATING #-} -- TODO: we should eliminate the tree as a traversable instead of doing it like this
render-element : Element → String
render-element (textnode x) = encode (S.toList "</>") x -- we don't technically need to escape all these characters, but it's easier to do this
render-element (node tag inner) =
  "<" S.++ tag .name S.++ S.concat (L.map (λ attr → " " S.++ render-attr attr) (tag .attributes)) S.++ ">"
  S.++ S.concat (L.map render-element inner)
  S.++ "</" S.++ tag .name S.++ ">"
  where
    open HTMLTag

