open import Agda.Primitive renaming (Set to Type)

open import Data.Product
open import Data.Bool hiding (_<_)

import Data.String as S
import Data.List as L
open S using (String)
open L using (List)

open import Base

open import HTML

module HTML.Syntax where

module Attributes where
  open HTMLTag
  mkAttr : (name : String) → {T (valid-attribute-name name)} → HTMLTag → String → HTMLTag
  mkAttr attr-name {attr-name-is-valid} tag value =
    record tag
    { attributes = record
        { name = attr-name
        ; value = value
        ; name-is-valid = attr-name-is-valid
        }
        L.∷ tag .attributes
    ; name = tag .name
    ; name-is-valid = tag .name-is-valid
    }

  _id=_ = mkAttr "id"
  infixl 20 _id=_
  _class=_ = mkAttr "class"
  infixl 20 _class=_
  _style=_ = mkAttr "style"
  infixl 20 _style=_
  _lang=_ = mkAttr "lang"
  infixl 20 _lang=_
  _src=_ = mkAttr "src"
  infixl 20 _src=_
  _href=_ = mkAttr "href"
  infixl 20 _href=_
  _name=_ = mkAttr "name"
  infixl 20 _name=_
  _content=_ = mkAttr "content"
  infixl 20 _content=_

open Attributes hiding (mkAttr) public

module Tags where
  mkTag : (name : String) → {T (valid-tag-name name)} → HTMLTag
  mkTag name {name-is-valid} = record { name = name ; attributes = L.[]; name-is-valid = name-is-valid }

  html = mkTag "html"
  body = mkTag "body"
  head = mkTag "head"
  meta = mkTag "meta"
  title = mkTag "title"
  img = mkTag "img"
  p = mkTag "p"
  span = mkTag "span"
  footer = mkTag "footer"

  open import Data.Nat
  open import Relation.Binary.PropositionalEquality
  postulate
    h-is-valid : (x : ℕ) → T (valid-tag-name ("h" S.++ show x)) -- :3
  h : ℕ → HTMLTag
  h x = record
    { name = "h" S.++ show x
    ; attributes = L.[]
    ; name-is-valid = h-is-valid x
    }

open Tags hiding (mkTag) public

record DOMAble (T : Type) : Type where
  field
    endom : T → List Element
open DOMAble ⦃...⦄

instance
  TextnodeDOMAble : DOMAble String
  TextnodeDOMAble .endom x = textnode x L.∷ L.[]

  ElementDOMAble : DOMAble Element
  ElementDOMAble .endom e = e L.∷ L.[]

  ListDOMAble : {T : Type} → ⦃ DOMAble T ⦄ → DOMAble (List T)
  ListDOMAble .endom x = L.concatMap endom x

  PairDOMAble : {T₀ T₁ : Type} → ⦃ DOMAble T₀ ⦄ → ⦃ DOMAble T₁ ⦄ → DOMAble (T₀ × T₁)
  PairDOMAble .endom (x , y) = endom x L.++ endom y

module Syntax where
  open HTMLTag
  <_>_</_> : {D : Type} → ⦃ DOMAble D ⦄ → (t₀ : HTMLTag) → D → (t₁ : HTMLTag) → {T (t₀ .name == t₁ .name)} → Element
  < t₀ > sub </ t₁ > = node t₀ (endom sub)

  <_/> : HTMLTag → Element
  <_/> t = node t L.[]
open Syntax public
