open import Agda.Primitive renaming (Set to Type)

import Data.String as S
open import Data.Nat
open import Data.Bool

open import Base

module HTTP.URL where

record Authority : Type where
  constructor _꞉_
  field
    host : S.String -- TODO: Split out IP:s as a separate type?
    port : ℕ
    -- TODO: maybe include {port-in-range} : port < 2 ^ 16
open Authority public

instance
  ShowAuthority : Show Authority
  ShowAuthority .show x =
    if x .port == 80
      then x .host
      else x .host S.++ ":" S.++ show (x .port)

infix 10 _꞉_

data Path : Type where
  $ : Path -- end
  _/_ : S.String → Path → Path
open Path public

instance
  ShowPath : Show Path
  ShowPath .show $ = "(empty path)"
  ShowPath .show (p / $) = p
  ShowPath .show (p / rest@(_ / _)) = p S.++ "/" S.++ show rest

  EqPath : Eq Path
  EqPath ._==_ $        $        = true
  EqPath ._==_ (p / ps) (q / qs) = p == q ∧ ps == qs
  EqPath ._==_ (_ / _)  $        = false
  EqPath ._==_ $        (_ / _)  = false

infixr 30 _/_

record URL : Type where
  constructor http://_/_¿_
  field
    authority : Authority
    path : Path
    query : Maybe S.String
open URL public

infix 0 http://_/_¿_

instance
  ShowURL : Show URL
  ShowURL .show url =
    "http://" S.++ show (url .authority) S.++ "/" S.++ show (url .path) S.++ ((λ q → "?" S.++ q) <$> url .query or-else "")

private
  sample-url : URL
  sample-url = http:// "coral.shoes" ꞉ 80 / ("pages" / "index.html" / $) ¿ real "key=value"

