{-# OPTIONS --postfix-projections #-}

module Indexed where

open import Relation.Binary.PropositionalEquality hiding ([_])
open import Data.Nat renaming (_≡ᵇ_ to _==ⁿ_; _<ᵇ_ to _<ⁿ_)
open import Data.Nat.Properties using (≤-refl; ≤-step; ≤-trans)
open import Data.Nat.DivMod hiding (result)
open import Data.Vec hiding ([_]; _>>=_; foldr; foldl) renaming (map to mapᵥ)
open import Data.List using (List; length) renaming ([] to []ₗ; _∷_ to _∷ₗ_; _++_ to _++ₗ_)
open import Data.Bool hiding (_<_; _≤_) renaming (_∧_ to _&&_; _∨_ to _||_; T to is-true)
open import Data.Product hiding (map; map₁; map₂)
open import Data.Unit using (⊤; tt)
open import Data.Empty using (⊥)

open import Agda.Primitive renaming (Set to Type)

open import Base
open import NonEmpty

module boolean-compare where
  if′_then_else_ : {A : Type} → (b : Bool) → (is-true b → A) → (is-false b → A) → A
  if′ true then a else _ = a tt
  if′ false then _ else b = b tt

  _≤ⁿ_ : ℕ → ℕ → Bool
  0 ≤ⁿ y = true
  suc x ≤ⁿ y = x <ⁿ y

  _>ⁿ_ : ℕ → ℕ → Bool
  a >ⁿ b = b <ⁿ a

  _≥ⁿ_ : ℕ → ℕ → Bool
  a ≥ⁿ b = b ≤ⁿ a

  <ⁿ-trans : (a b c : ℕ) → is-true (a <ⁿ b) → is-true (b <ⁿ c) → is-true (a <ⁿ c)
  <ⁿ-trans zero    (suc b) (suc c) tt sb<sc = tt
  <ⁿ-trans (suc a) (suc b) (suc c) sa<sb sb<sc = <ⁿ-trans a b c sa<sb sb<sc

  -- needed for conversion because DivMod uses nat's comparisons and we want to use them
  <→<ⁿ : {a b : ℕ} → a < b → is-true (a <ⁿ b)
  <→<ⁿ {zero} {suc b} (s≤s z≤n) = tt
  <→<ⁿ {suc a} {suc b} (s≤s sa<sb) = <→<ⁿ {a} {b} sa<sb

  ≤→≤ⁿ : {a b : ℕ} → a ≤ b → is-true (a ≤ⁿ b)
  ≤→≤ⁿ {zero} {b} a≤b = tt
  ≤→≤ⁿ {suc a} {b} a≤b = <→<ⁿ {a} {b} a≤b

  <ⁿ→< : {a b : ℕ} → is-true (a <ⁿ b) → a < b
  <ⁿ→< {zero} {suc b} t = s≤s z≤n
  <ⁿ→< {suc a} {suc b} t = s≤s (<ⁿ→< {a} {b} t)

  ≤ⁿ→≤ : {a b : ℕ} → is-true (a ≤ⁿ b) → a ≤ b
  ≤ⁿ→≤ {zero} {b} t = z≤n
  ≤ⁿ→≤ {suc a} {b} t = <ⁿ→< {a} {b} t
open boolean-compare public

k′ : Type → (ℕ → Type)
(k′ A) i = A

-- lift functions
_→′_ : (ℕ → Type) → (ℕ → Type) → (ℕ → Type)
(A →′ B) i = A i → B i

-- lift products
_×′_ : (ℕ → Type) → (ℕ → Type) → (ℕ → Type)
(A ×′ B) i = A i × B i

-- lower to implicit
[_] : (ℕ → Type) → Type
[ A ] = {i : ℕ} → A i

record □_ (A : ℕ → Type) (n : ℕ) : Type where
  constructor mk□
  field
    call : {m : ℕ} → m < n → A m
open □_ public

□-0 : {A : ℕ → Type} → (□ A) 0
□-0 .call ()

infix 0 [_]
infixr 1 _×′_
infixr 1 _→′_
infixl 100 □_

map : {A B : ℕ → Type} → [ A →′ B ] → [ □ A →′ □ B ]
map f □a .call lt = f (□a .call lt)

module map-elaborate where
  map₁ map₂ : {A B : ℕ → Type}
      → ({i : ℕ} →    A  i →    B  i)
      → ({n : ℕ} → (□ A) n → (□ B) n)
  map₁ f {n} □a .call {m} m<n = f (□a .call {m} m<n)
  map₂ f = λ □a → mk□ λ m<n → f (□a .call m<n)

-- we have for all n, for all m<n, A
-- specialize n = suc i, supply i < suc i
extract : {A : ℕ → Type} → [ □ A ] → [ A ]
extract □a {i} = □a .call {i} (≤-refl {suc i})

enbox : {A : ℕ → Type} → [ A ] → [ □ A ]
enbox a = mk□ λ {m} pf → a {m}

strong-induction : {A : ℕ → Type}
                 → ((n : ℕ) → ((m : ℕ) → m < n → A m) → A n)
                 → (n : ℕ)
                 → A n
strong-induction {A} ind = λ n → helper (suc n) n ≤-refl
  where
    data _∨_ (A B : Type) : Type where
      inl : A → A ∨ B
      inr : B → A ∨ B

    lt-split : (a b : ℕ) → a < suc b → (a < b) ∨ (a ≡ b)
    lt-split zero zero p = inr refl
    lt-split zero (suc b) p = inl (s≤s z≤n)
    lt-split (suc a) zero (s≤s ())
    lt-split (suc a) (suc b) (s≤s p) with lt-split a b p
    ... | inl a<b = inl (s≤s a<b)
    ... | inr a≡b = inr (cong suc a≡b)

    helper : (n : ℕ) → (m : ℕ) → m < n → A m
    helper zero m ()
    helper n@(suc pn) m m<n with lt-split m pn m<n
    ... | inl m<pn = helper pn m m<pn
    ... | inr m≡pn = ind m (subst (λ a → (m′ : ℕ) → m′ < a → A m′) (sym m≡pn) (helper pn))

ℕ-ind : {P : ℕ → Type}
      → ((n : ℕ) → ((m : ℕ) → @0 m < n → P m) → P n)
      → (n : ℕ)
      → P n
ℕ-ind {P} f n = f n (go n ≤-refl)
  where
    -- m < p ≤ n
    go : (p : ℕ) → @0 p ≤ n → (m : ℕ) → @0 m < p → P m
    go _ _ zero _ = f zero (λ m ())
    go p@(suc pp) (s≤s l) m@(suc pm) (s≤s q) = f m λ k x → go pp (≤-step l) k (≤-trans x q)

fix : {A : ℕ → Type}
    → [ □ A →′ A ]
    → [ A ]
--fix {A} □a→a {n} = strong-induction (λ n ind → □a→a (mk□ λ {m} m<n → ind m m<n)) n
fix {A} □a→a {n} = ℕ-ind (λ n ind → □a→a (mk□ λ {m} m<n → ind m m<n)) n
