{-# OPTIONS --guardedness --postfix-projections #-}

module main where

open import Agda.Primitive renaming (Set to Type)

import Data.String as S
import Data.Vec as V
import Data.List as L
open S using (String)
open L using (List)
open V using (Vec)

open import Data.Product using (_×_; proj₁; proj₂; _,_)
open import Data.Nat hiding (_<_; _>_)
open import Data.Unit using (⊤; tt)
open import Data.Bool

open import Base
open import Bits-and-Bytes using (Byte; ShowByteList)
open import SysIO
open import Socket
open import Indexed
open import Parsing (Byte)
open import NonEmpty

import HTTP
import HTTP.Parse
import HTTP.URL as URL
open import HTML
import HTML.Syntax
import HTML.Render

import UTF-8
import ASCII

module Pages where
  open HTML

  render-index : Maybe S.String → HTML.Element
  render-index query =
    < html lang= "en" style= "color: white; background: black; padding: 1em;" >
      < head >
        < title > "henttp index" </ title > ,
        < meta name= "description" content= "Simple webpage rendered in Agda DSL" /> ,
        < meta name= "author" content= "xenia" />
      </ head > ,
      < body >
        < h 1 >
          < span style= "color: yellow;" > "hen" </ span > ,
          < span style= "color: fuchsia;" > "ttp" </ span >
        </ h 1 > ,
        < h 3 > "what is this?" </ h 3 > ,
        < p >
          "henttp is an HTTP server, HTML DSL and (soon to be) router, written in Agda"
        </ p >
      </ body > ,
      < footer >
        "meow"
      </ footer >
    </ html >
    where
      open HTML.Syntax

handle-request : HTTP.Request → IO HTTP.Response
handle-request req =
  case req .target .path of λ where
    ($) →
      pure (
        record
          { status-code = 302
          ; status = "Found"
          ; headers =
            record
              { name = ASCII.encode-list⁺ "Location"
              ; value = ASCII.encode-list⁺ ("/" S.++ show ("index.html" / $))
              } L.∷ L.[]
          ; content = cake
          }
      )
    ("index.html" / $) →
      pure (
        HTTP.200-ok HTTP.text/html (HTML.Render.render-element (Pages.render-index (req .target .query)))
      )
    _ → pure HTTP.404-not-found
  where
    open HTTP.Request using (target)
    open URL
    open import ParseHelpers using (number)


handle : Socket → IO (List Byte)
handle sock = do
  putStrLn "handle: new connection‼"
  let got = get-bytes sock

  case HTTP.Parse.parse-request .parse (V.fromList got) of λ where
    (real ⟨ req , _ , _ ⟩) →
      (do
        putStrLn "handle: got request"
        putStrLn ("  method = " S.++ show (req .method))
        putStrLn ("  target = " S.++ show (req .target))
        putStrLn ("  version = " S.++ show (req .version))
        putStrLn ("  headers = " S.++ show (req .headers))
        case req .content of λ where
          (real c) → putStrLn ("  content = " S.++ show ⦃ ShowByteList ⦄ c)
          cake → putStrLn "  no content"
        HTTP.encode-response <$> handle-request req
      )
    cake →
      (do
        putStrLn "Got an invalid request"
        HTTP.encode-response <$> pure HTTP.400-bad-request
      )
  where
    open HTTP.Request

run-port : ℕ
run-port = 1337

main : IO ⊤
main = do
  putStrLn "main: starting"
  putStrLn ("main: running on port " S.++ show run-port)
  run-server "localhost" run-port handle
