open import Agda.Primitive renaming (Set to Type)

open import Data.Product
open import Data.Nat
open import Data.List using (List; []; _∷_)
open import Data.Bool

import Data.String as S
import Data.List as L
import Data.Vec as V
open S using (String)
open L using (List)
open V using (Vec)

open import Base

module NonEmpty where

data List⁺ (A : Type) : Type where
  [_]⁺ : A → List⁺ A
  _∷⁺_ : A → List⁺ A → List⁺ A

infixr 0 _∷⁺_

foldr : {A B : Type} → (A → B → B) → B → List⁺ A → B
foldr _+_ base [ x ]⁺ = x + base
foldr _+_ base (x ∷⁺ xs) = x + foldr _+_ base xs

foldl : {A B : Type} → (B → A → B) → B → List⁺ A → B
foldl _+_ acc [ x ]⁺ = acc + x
foldl _+_ acc (x ∷⁺ xs) = foldl _+_ (acc + x) xs

cons⁺ : {A : Type} → A → List A → List⁺ A
cons⁺ a [] = [ a ]⁺
cons⁺ a (x ∷ xs) = a ∷⁺ cons⁺ x xs

_⁺++⁺_ : {A : Type} → List⁺ A → List⁺ A → List⁺ A
[ x ]⁺ ⁺++⁺ l₁ = x ∷⁺ l₁
(x ∷⁺ l₀) ⁺++⁺ l₁ = x ∷⁺ (l₀ ⁺++⁺ l₁)

_++⁺_ : {A : Type} → List A → List⁺ A → List⁺ A
[] ++⁺ l₁ = l₁
(x ∷ l₀) ++⁺ l₁ = x ∷⁺ (l₀ ++⁺ l₁)

_⁺++_ : {A : Type} → List⁺ A → List A → List⁺ A
[ x ]⁺ ⁺++ [] = [ x ]⁺
[ x ]⁺ ⁺++ (y ∷ l₁) = (x ∷⁺ [ y ]⁺ ) ⁺++ l₁
(x ∷⁺ l₀) ⁺++ l₁ = x ∷⁺ (l₀ ⁺++ l₁)

infixr 20 _⁺++⁺_ _++⁺_ _⁺++_

list⁺-to-list : {A : Type} → List⁺ A → List A
list⁺-to-list [ x ]⁺ = x ∷ []
list⁺-to-list (x ∷⁺ xs) = x ∷ list⁺-to-list xs

list⁺-to-vec : {A : Type} → List⁺ A → Σ ℕ λ n → V.Vec A (suc n)
list⁺-to-vec [ x ]⁺ = 0 , x V.∷ V.[]
list⁺-to-vec (x ∷⁺ xs) = _ , x V.∷ (proj₂ (list⁺-to-vec xs))

vec-to-list⁺ : {A : Type} → {n : ℕ} → V.Vec A (suc n) → List⁺ A
vec-to-list⁺ (x V.∷ V.[]) = [ x ]⁺
vec-to-list⁺ (x V.∷ xs@(_ V.∷ _)) = x ∷⁺ vec-to-list⁺ xs

list-to-list⁺? : {A : Type} → List A → Maybe (List⁺ A)
list-to-list⁺? [] = cake
list-to-list⁺? (x ∷ xs) = real (cons⁺ x xs)


instance
  ShowList⁺ : {A : Type} → ⦃ Show A ⦄ → Show (List⁺ A)
  ShowList⁺ .show x = show (list⁺-to-list x) S.++ "⁺"
    where
      import Data.String as String

  EqList⁺ : {T : Type} → ⦃ Eq T ⦄ → Eq (List⁺ T)
  EqList⁺ ._==_ [ x ]⁺ [ y ]⁺       = x == y
  EqList⁺ ._==_ (x ∷⁺ xs) (y ∷⁺ ys) = x == y ∧ xs == ys
  EqList⁺ ._==_ (_ ∷⁺ _)  [ _ ]⁺    = false
  EqList⁺ ._==_ [ _ ]⁺    (_ ∷⁺ _)  = false
