open import Agda.Primitive renaming (Set to Type)

open import Relation.Binary.PropositionalEquality hiding ([_])

open import Data.Nat
open import Data.Bool hiding (_<_)
open import Data.Unit
open import Data.Product

import Data.List as L
import Data.Vec as V
import Data.String as S
import Data.Char as C
open L using (List)
open V using (Vec)
open S using (String)
open C using (Char)

open import Base
open import Indexed
open import NonEmpty
open import Bits-and-Bytes
import UTF-8

open import Parsing (Byte)

module ParseHelpers where

-- Bytes are compared against codepoints, meaning any non-ascii Chars will be death
_is_ : Byte → Char → Bool
x is c = x .value == C.toℕ c

_between_and_ : Byte → Char → Char → Bool
x between a and b = (C.toℕ a ≤ᵇ x .value) ∧ (x .value ≤ᵇ C.toℕ b)


any-of : String → [ Parser Byte ]
any-of x = L.foldr (λ ch p → ((_is ch) <?>′ any₁) <|>′ p) fail (S.toList x)

uppercase lowercase letter digit : [ Parser Byte ]
uppercase = (_between 'A' and 'Z') <?>′ any₁
lowercase = (_between 'a' and 'z') <?>′ any₁
digit     = (_between '0' and '9') <?>′ any₁
letter = uppercase <|>′ lowercase

cr lf crlf : [ Parser ⊤ ]
cr = (λ _ → tt) <$>′ any-of "\r"
lf = (λ _ → tt) <$>′ any-of "\n"
crlf = (cr >>′ enbox lf) <|>′ lf -- we allow lf because :3

space spaces : [ Parser ⊤ ]
space = (λ _ → tt) <$>′ any-of " "
spaces = (λ _ → tt) <$>′ many space

percent-encoded : [ Parser Byte ]
percent-encoded = (λ where (a V.∷ b V.∷ V.[]) → mkClip (a * 16 + b)) <$>′ unencoded-percent-encoding
  where
    unhex : Byte → Maybe ℕ -- char to hex digit
    unhex x =
      if x between '0' and '9'
         then real (x .value ∸ (C.toℕ '0'))
         else if x between 'a' and 'f'
           then real (x .value ∸ (C.toℕ 'a') + 10)
           else if x between 'A' and 'F'
             then real (x .value ∸ (C.toℕ 'A') + 10)
             else cake

    hex-digit : [ Parser ℕ ]
    hex-digit = unhex <$?>′ any₁

    unencoded-percent-encoding : [ Parser (Vec ℕ 2) ]
    unencoded-percent-encoding = exact (bytes.PERCENT V.∷ V.[]) >>′ enbox (repeat 2 hex-digit)

digit-ℕ : [ Parser ℕ ]
digit-ℕ = (λ x → x .value ∸ C.toℕ '0') <$>′ digit

number : [ Parser ℕ ]
number = foldl (λ n d → n * 10 + d) 0 <$>′ many digit-ℕ
