open import Data.Nat renaming (_≡ᵇ_ to _==ⁿ_; _<ᵇ_ to _<ⁿ_)
open import Data.Nat.Properties using (≤-refl; <-trans; 0<1+n)
open import Data.Nat.DivMod hiding (result)
open import Data.Vec hiding ([_]; _>>=_; foldr; foldl) renaming (map to mapᵥ)
open import Data.List using (List; length) renaming ([] to []ₗ; _∷_ to _∷ₗ_; _++_ to _++ₗ_)
open import Data.Bool hiding (_<_; _≤_) renaming (_∧_ to _&&_; _∨_ to _||_; T to is-true)
open import Data.Product hiding (map; map₁; map₂)
open import Data.Unit using (⊤; tt)
open import Data.Empty using (⊥)

open import Agda.Primitive renaming (Set to Type)

open import Base
open import NonEmpty
open import Indexed

module Parsing (T : Type) ⦃ EqT : Eq T ⦄ where

record Parse (A : Type) (n : ℕ) : Type where
  constructor ⟨_,_,_⟩
  field
    result : A
    {m} : ℕ
    m<n : m < n
    rest : Vec T m
open Parse public

record Parser (A : Type) (n : ℕ) : Type where
  field
    parse : Vec T n → Maybe (Parse A n)

open Parser public

parse-list⁺ : {A : Type} → [ Parser A ] → List⁺ T → Maybe A
parse-list⁺ p l =
  p .parse (proj₂ (list⁺-to-vec l))
  >>= λ parse →
    if parse .m == 0
      then pure (parse .result)
      else cake

-- Parser combinators
_<$>′_ : {A B : Type} → (A → B) → [ Parser A →′ Parser B ]
(f <$>′ p) .parse a = do
  ⟨ x , m<n , rest ⟩ ← p .parse a
  pure ⟨ f x , m<n , rest ⟩

_<$?>′_ : {A B : Type} → (A → Maybe B) → [ Parser A →′ Parser B ]
(f <$?>′ p) .parse a = do
  ⟨ x , m<n , rest ⟩ ← p .parse a
  y ← f x
  pure ⟨ y , m<n , rest ⟩

_<?>′_ : {A : Type} → (A → Bool) → [ Parser A →′ Parser A ]
pred <?>′ p = (λ a → if pred a then real a else cake) <$?>′ p

_<|>′_ : {A : Type} → [ Parser A →′ Parser A →′ Parser A ]
(p₀ <|>′ p₁) .parse a = case p₀ .parse a of λ where
  (real x) → real x
  cake → p₁ .parse a

infixr 20 _<|>′_

_<&>′_ : {A B : Type} → [ Parser A →′ □ (Parser B) →′ Parser (A × B) ]
(p₀ <&>′ p₁) .parse a = do
  ⟨ x₀ , m<i , rest ⟩ ← p₀ .parse a
  ⟨ x₁ , n<m , rest ⟩ ← (p₁ .call m<i) .parse rest
  pure ⟨ (x₀ , x₁) , <-trans n<m m<i , rest ⟩

_<&⃖>′_ : {A B : Type} → [ Parser A →′ □ (Parser B) →′ Parser A ]
(p₀ <&⃖>′ p₁) .parse a = do
  ⟨ x₀ , m<i , rest ⟩ ← p₀ .parse a
  ⟨ _ , n<m , rest ⟩ ← (p₁ .call m<i) .parse rest
  pure ⟨ x₀ , <-trans n<m m<i , rest ⟩

_<&⃗>′_ : {A B : Type} → [ Parser A →′ □ (Parser B) →′ Parser B ]
(p₀ <&⃗>′ p₁) .parse a = do
  ⟨ _ , m<i , rest ⟩ ← p₀ .parse a
  ⟨ x₁ , n<m , rest ⟩ ← (p₁ .call m<i) .parse rest
  pure ⟨ x₁ , <-trans n<m m<i , rest ⟩

_<&?>′_ : {A B : Type} → [ Parser A →′ □ (Parser B) →′ Parser (A × Maybe B) ]
(p₀ <&?>′ p₁) .parse a = do
  ⟨ x₀ , m<i , rest ⟩ ← p₀ .parse a
  case (p₁ .call m<i) .parse rest of λ where
    (real ⟨ x₁ , n<m , rest ⟩) → pure ⟨ (x₀ , real x₁) , <-trans n<m m<i , rest ⟩
    cake → pure ⟨ (x₀ , cake) , m<i , rest ⟩

_<&>□_ : {A B : Type} → [ Parser A ] → [ Parser B ] → [ Parser (A × B) ]
p₀ <&>□ p₁ = p₀ <&>′ enbox p₁

_<&⃖>□_ : {A B : Type} → [ Parser A ] → [ Parser B ] → [ Parser A ]
p₀ <&⃖>□ p₁ = p₀ <&⃖>′ enbox p₁

_<&⃗>□_ : {A B : Type} → [ Parser A ] → [ Parser B ] → [ Parser B ]
p₀ <&⃗>□ p₁ = p₀ <&⃗>′ enbox p₁

_<&?>□_ : {A B : Type} → [ Parser A ] → [ Parser B ] → [ Parser (A × Maybe B) ]
p₀ <&?>□ p₁ = p₀ <&?>′ enbox p₁

infix 20 _<&>′_
infix 20 _<&⃗>′_
infix 20 _<&⃖>′_
infix 20 _<&?>′_

infixr 20 _<&>□_
infixr 20 _<&⃗>□_
infixr 20 _<&⃖>□_
infixr 20 _<&?>□_

many-fix : {A : Type} → [ Parser A ] → [ Parser (List⁺ A) ]
many-fix p = fix λ x → ((λ (a , b) → a ∷⁺ b) <$>′ (p <&>′ x))
                       <|>′ ([_]⁺ <$>′ p)

-- many has the exact same definition as many-fix
-- as many-fix is terminating, this must be as well
-- :3
{-# TERMINATING #-}
many-recursing : {A : Type} → [ Parser A ] → [ Parser (List⁺ A) ]
many-recursing p = ((λ (a , b) → a ∷⁺ b) <$>′ (p <&>′ enbox (many-recursing p)))
                   <|>′ ([_]⁺ <$>′ p)

many : {A : Type} → [ Parser A ] → [ Parser (List⁺ A) ]
many = many-recursing

upto : {A : Type} → (n : ℕ) → {is-true (0 <ⁿ n)} → [ Parser A ] → [ Parser (List⁺ A) ]
upto {A} (suc pn) p = go pn
  where
    go : ℕ → [ Parser (List⁺ A) ]
    go 0 = [_]⁺ <$>′ p
    go n@(suc pn) =
       ((λ (x , xs) → x ∷⁺ xs) <$>′ (p <&>′ enbox upto-pn))
       <|>′ upto-pn
      where
        upto-pn : [ Parser (List⁺ A) ]
        upto-pn = go pn

_>>=′_ : {A B : Type} → [ Parser A →′ ((k′ A) →′ □ Parser B) →′ Parser B ]
(pa >>=′ f) .parse inp = do
  ⟨ a , lt , inp′ ⟩ ← pa .parse inp
  let pb = f a
  ⟨ b , lt′ , inp′′ ⟩ ← (pb .call lt) .parse inp′
  pure ⟨ b , <-trans lt′ lt , inp′′ ⟩

_>>′_ : {A B : Type} → [ Parser A →′ □ Parser B →′ Parser B ]
pa >>′ pb = pa >>=′ λ _ → pb

_>$=′_ : {A B : Type} → [ Parser A →′ k′ (A → B) →′ Parser B ]
pa >$=′ f = f <$>′ pa

_>?=′_ : {A : Type} → [ Parser A →′ k′ (A → Bool) →′ Parser A ]
pa >?=′ f = f <?>′ pa

_>$?=′_ : {A B : Type} → [ Parser A →′ k′ (A → Maybe B) →′ Parser B ]
pa >$?=′ f = f <$?>′ pa

infixl 10 _>>=′_ _>>′_ _>$=′_ _>$?=′_ _>?=′_

fail : {A : Type} → [ Parser A ]
fail .parse _ = cake

-- simple parsers

any₁ : [ Parser T ]
any₁ .parse [] = cake
any₁ .parse (x ∷ xs) = real ⟨ x , ≤-refl , xs ⟩

repeat : {A : Type} → (n : ℕ) → .{nonempty : is-true (0 <ⁿ n)} → [ Parser A ] → [ Parser (Vec A n) ]
repeat 1 p = (λ x → x ∷ []) <$>′ p
repeat n@(suc pn@(suc _)) p =
  p >>=′ λ x →
  enbox ((λ xs → x ∷ xs) <$>′ repeat pn p)

exact : {n : ℕ} → Vec T (suc n) → [ Parser ⊤ ]
exact {n} x = go x
  where
    check : T → T → Maybe ⊤
    check x y = if x == y then real tt else cake

    go : {n : ℕ} → .{is-true (0 <ⁿ n)} → Vec T n → [ Parser ⊤ ]
    go (x ∷ []) = check x <$?>′ any₁
    go {n@(suc pn)} (x ∷ xs@(_ ∷ _)) =
      (check x <$?>′ any₁) >>=′ λ _ → enbox (go xs)

{-
module parse-example where
  open import Data.String
  open import Data.Char renaming (_==_ to _==ᶜ_)

  open parse (Char) (_==ᶜ_)

  parse-digit : [ Parser ℕ ]
  parse-digit = (decode-char <$?> any₁)
    where
      decode-char : Char → Maybe ℕ
      decode-char '0' = real 0
      decode-char '1' = real 1
      decode-char '2' = real 2
      decode-char '3' = real 3
      decode-char '4' = real 4
      decode-char '5' = real 5
      decode-char '6' = real 6
      decode-char '7' = real 7
      decode-char '8' = real 8
      decode-char '9' = real 9
      decode-char _ = cake

  parse-ℕ : [ Parser ℕ ]
  parse-ℕ = foldl (λ n d → n * 10 + d) 0 <$> some parse-digit

  sample : parse-ℕ .parse (toVec "123xy")
         ≡ real ⟨ 123 , _ , toVec "xy" ⟩
  sample = refl
-}
