{-# OPTIONS --guardedness --postfix-projections #-}
module Socket where

open import Agda.Primitive renaming (Set to Type)

open import Data.Product hiding (map)
open import Data.Nat
open import Data.String
open import Data.List
open import Data.Unit

open import Base
open import SysIO
open import Bits-and-Bytes

module SocketForeign where
  {-# FOREIGN GHC
    import qualified Data.Word as Word
    import qualified Data.Functor as Funct
    import qualified Control.Monad as Monad
    import qualified Data.Text as T
    import qualified System.IO as SIO
    import qualified Control.Concurrent.STM as STM
    import qualified Control.Concurrent.STM.TVar as TVar

    import qualified Data.ByteString as BS

    import System.IO.Unsafe as Unsafe

    import qualified Network.Socket as Socket
    import qualified Network.Socket.ByteString as SocketBS
    import qualified Network.Run.TCP as RunTCP
  #-}

  -- Data.Product doesn't do any FFI, and we can't define the FFI outside the module,
  -- so we define our own product type
  data Tuple2 (A B : Type) : Type where
    mkTuple2 : A → B → Tuple2 A B

  {-# FOREIGN GHC data Tuple2 a b = MKTuple2 a b #-}
  {-# COMPILE GHC Tuple2 = data Tuple2 (MKTuple2) #-}

  postulate
    RawSocket : Type
    raw-uncons : RawSocket → Maybe (Tuple2 Word8 RawSocket)

    run-raw-server : String → ℕ → (RawSocket → IO (List Word8)) → IO ⊤

  {-# FOREIGN GHC
    innerSockSize :: Int
    innerSockSize = 1 -- TODO: Increase

    debug :: Bool
    debug = False

    info :: Bool
    info = False

    gray :: String
    gray = "\x1b[38;5;8m"

    blue :: String
    blue = "\x1b[38;5;4m"

    reset :: String
    reset = "\x1b[0m"

    printDebug :: String -> String -> IO ()
    printDebug f x = Monad.when debug (SIO.putStrLn $ gray ++ "  [" ++ f ++ "] " ++ x ++ reset)

    printInfo :: String -> String -> IO ()
    printInfo f x = Monad.when info (SIO.putStrLn $ blue ++ "  [" ++ f ++ "] " ++ x ++ reset)

    data BufferedSocket = BufferedSocket
      { getSocket :: Socket.Socket
      , getBuffer :: TVar.TVar BS.ByteString
      , getOffset :: Int
      }

    mkBuff :: Socket.Socket -> IO BufferedSocket
    mkBuff s = do
      buf <- TVar.newTVarIO BS.empty
      pure BufferedSocket
        { getSocket = s
        , getBuffer = buf
        , getOffset = 0
        }

    fillBuffer :: BufferedSocket -> IO ()
    fillBuffer s = do
      remaining <- (TVar.readTVarIO $ getBuffer s) >>= (\buf -> pure $ innerSockSize - (BS.length buf) + getOffset s)
      printDebug "fillBuffer" $ "remaining = " ++ show remaining
      if remaining <= 0
        then pure ()
        else do
          printDebug "fillBuffer" $ "Reading " ++ show remaining ++ " bytes"
          got <- SocketBS.recv (getSocket s) remaining
          printInfo "fillBuffer" $ "got: " ++ show got
          STM.atomically $ TVar.modifyTVar (getBuffer s) (flip BS.append got)

    unconsIO :: BufferedSocket -> IO (Maybe (Word.Word8, BufferedSocket))
    unconsIO s = do
      buf <- TVar.readTVarIO $ getBuffer s
      printDebug "unconsIO" $ "buffer = " ++ show buf ++ ", offset = " ++ show (getOffset s)
      let next = s { getOffset = 1 + getOffset s }
      let offset = getOffset s
      if offset < BS.length buf
        then pure $ pure (BS.index buf offset, next)
        else do
          printDebug "unconsIO" $ "filling buffer"
          fillBuffer s
          buf <- TVar.readTVarIO $ getBuffer s
          printDebug "unconsIO" $ "after: buffer = " ++ show buf ++ ", offset = " ++ show (getOffset s)
          if offset < BS.length buf
            then pure $ pure (BS.index buf offset, next)
            else pure Nothing
  #-}

  {-# COMPILE GHC RawSocket = type BufferedSocket #-}
  {-# COMPILE GHC raw-uncons = \s -> (\(x, y) -> MKTuple2 x y) Funct.<$> Unsafe.unsafePerformIO (unconsIO s) #-}

  {-# COMPILE GHC run-raw-server =
      \ hostname port runFn ->
        RunTCP.runTCPServer
          (Just (T.unpack hostname))
          (show port)
          (\socket -> do
            sock <- mkBuff socket ;
            dat <- (BS.singleton <$>) <$> runFn sock ;
            Monad.forM_ dat (\b -> printInfo "run-raw-server" ("sending " ++ show b) >> SocketBS.sendAll socket b)
          )
  #-}
open SocketForeign using (RawSocket; raw-uncons; run-raw-server)

record Socket : Type where
  coinductive
  field
    split : Maybe (Byte × Socket)
open Socket public

module SocketTrans where
  unraw : RawSocket → Socket
  unraw raw-sock .split with raw-uncons raw-sock
  ...                   | real ( SocketForeign.mkTuple2 w rest ) = real ( Word8-to-Byte w , unraw rest)
  ...                   | cake = cake

run-server : String → ℕ → (Socket → IO (List Byte)) → IO ⊤
run-server hostname port f = run-raw-server hostname port (λ rs → map Byte-to-Word8 <$> (f (SocketTrans.unraw rs)))

{-# TERMINATING #-} -- hopefully the user hangs up at some point :)
get-bytes : Socket → List Byte
get-bytes sock with sock .split
...            | real (ch , sock′) = ch ∷ get-bytes sock′
...            | cake = []
