module SysIO where

open import Agda.Primitive renaming (Set to Type)

open import Data.String using (String; _++_)
open import Data.Unit using (⊤; tt)

open import Base
open import Bits-and-Bytes

module IOForeign where
  postulate
    IO : Type → Type
    io-pure : {A : Type} → A → IO A
    io-bind : {A B : Type} → IO A → (A → IO B) → IO B

    putStr : String → IO ⊤

  {-# BUILTIN IO IO #-}
  {-# FOREIGN GHC
    import qualified Control.Monad as Monad
    import qualified Control.Applicative as Applicative

    import qualified System.IO as SIO
    import qualified Data.Text as T
    import qualified Data.Text.IO as TIO
  #-}
  {-# COMPILE GHC IO = type IO #-}
  {-# COMPILE GHC io-pure = \ty a -> Applicative.pure a #-}
  {-# COMPILE GHC io-bind = \tya tyb a f -> (Monad.>>=) a f #-}
  {-# COMPILE GHC putStr = TIO.putStr #-}

open IOForeign using (IO; putStr) public

instance
  MonadIO : Monad IO
  MonadIO .pure = IOForeign.io-pure
  MonadIO ._>>=_ = IOForeign.io-bind

putStrLn : String → IO ⊤
putStrLn x = putStr (x ++ "\n")
