{-# OPTIONS --postfix-projections #-}
module UTF-8 where

open import Agda.Primitive renaming (Set to Type)

open import Data.Product
open import Data.Unit

open import Relation.Binary.PropositionalEquality hiding ([_])
open import Data.Bool hiding (_<_; _<?_)
open import Data.Nat
open import Data.Nat.Properties using (<-trans; <ᵇ⇒<; ≤-refl; m≤n+m)
open import Relation.Nullary

import Data.Vec as V
import Data.List as L
import Data.String as S
import Data.Char as C
open V using (Vec)
open L using (List)
open S using (String)
open C using (Char)

open import Base
open import NonEmpty
open import Bits-and-Bytes

postulate
  char-toℕ-in-range : (c : C.Char) → C.toℕ c < 0x110000

encode-char : C.Char → List Byte
encode-char ch = case x <? 0x80 of λ where
    (yes x<0x80) → encode-1 x x<0x80
    (no _) → case x <? 0x800 of λ where
      (yes x<0x800) → encode-2 x x<0x800
      (no _) → case x <? 0x10000 of λ where
        (yes x<0x10000) → encode-3 x x<0x10000
        (no _) → encode-4 x x<0x110000
  where
    open V

    x : ℕ
    x = C.toℕ ch

    x<0x110000 : x < 0x110000
    x<0x110000 = char-toℕ-in-range ch

    encode-1 : (n : ℕ) → n < 0x80 → List Byte
    encode-1 n n<0x80 = case enbits {7} n n<0x80 of λ where
      (b0 ∷ b1 ∷ b2 ∷ b3 ∷ b4 ∷ b5 ∷ b6 ∷ []) →
        Bit8-to-Byte (b0 ∷ b1 ∷ b2 ∷ b3 ∷ b4 ∷ b5 ∷ b6 ∷ 𝟘 ∷ []) L.∷
        L.[]

    encode-2 : (n : ℕ) → n < 0x800 → List Byte
    encode-2 n n<0x800 = case enbits {11} n n<0x800 of λ where
      (b0 ∷ b1 ∷ b2 ∷ b3 ∷ b4 ∷ b5 ∷ b6 ∷ b7 ∷ b8 ∷ b9 ∷ b10 ∷ []) →
        Bit8-to-Byte (b6 ∷ b7 ∷ b8 ∷ b9 ∷ b10 ∷ 𝟘  ∷ 𝟙 ∷ 𝟙 ∷ []) L.∷
        Bit8-to-Byte (b0 ∷ b1 ∷ b2 ∷ b3 ∷ b4  ∷ b5 ∷ 𝟘 ∷ 𝟙 ∷ []) L.∷
        L.[]

    encode-3 : (n : ℕ) → n < 0x10000 → List Byte
    encode-3 n n<0x10000 = case enbits {16} n n<0x10000 of λ where
      (b0 ∷ b1 ∷ b2 ∷ b3 ∷ b4 ∷ b5 ∷ b6 ∷ b7 ∷ b8 ∷ b9 ∷ b10 ∷ b11 ∷ b12 ∷ b13 ∷ b14 ∷ b15 ∷ []) →
        Bit8-to-Byte (b12 ∷ b13 ∷ b14 ∷ b15 ∷ 𝟘   ∷ 𝟙   ∷ 𝟙 ∷ 𝟙 ∷ []) L.∷
        Bit8-to-Byte (b6  ∷ b7  ∷ b8  ∷ b9  ∷ b10 ∷ b11 ∷ 𝟘 ∷ 𝟙 ∷ []) L.∷
        Bit8-to-Byte (b0  ∷ b1  ∷ b2  ∷ b3  ∷ b4  ∷ b5  ∷ 𝟘 ∷ 𝟙 ∷ []) L.∷
        L.[]

    encode-4 : (n : ℕ) → n < 0x110000 → List Byte
    encode-4 n n<0x110000 = case enbits {21} n (<-trans n<0x110000 (<ᵇ⇒< 0x110000 (2 ^ 21) tt)) of λ where
      (b0  ∷ b1  ∷ b2  ∷ b3  ∷ b4  ∷ b5 ∷ b6 ∷ b7 ∷ b8 ∷ b9 ∷ b10 ∷ b11 ∷ b12 ∷ b13 ∷ b14 ∷ b15 ∷
       b16 ∷ b17 ∷ b18 ∷ b19 ∷ b20 ∷ []) →
        Bit8-to-Byte (b18 ∷ b19 ∷ b20 ∷ 𝟘   ∷ 𝟙   ∷ 𝟙   ∷ 𝟙 ∷ 𝟙 ∷ []) L.∷
        Bit8-to-Byte (b12 ∷ b13 ∷ b14 ∷ b15 ∷ b16 ∷ b17 ∷ 𝟘 ∷ 𝟙 ∷ []) L.∷
        Bit8-to-Byte (b6  ∷ b7  ∷ b8  ∷ b9  ∷ b10 ∷ b11 ∷ 𝟘 ∷ 𝟙 ∷ []) L.∷
        Bit8-to-Byte (b0  ∷ b1  ∷ b2  ∷ b3  ∷ b4  ∷ b5  ∷ 𝟘 ∷ 𝟙 ∷ []) L.∷
        L.[]

encode-string : String → List Byte
encode-string s = go (S.toList s)
  where
    go : List Char → List Byte
    go L.[] = L.[]
    go (c L.∷ cs) = encode-char c L.++ go cs

module decode where
  open import Indexed using ([_]; _→′_; □_; fix)
  open import Parsing (Byte)
  import Data.Fin as F

  module _ where
    open V hiding ([_])
    decode-≤-1 : {m : ℕ} → Vec Byte (1 + m) → Maybe (Parse C.Char (1 + m))
    decode-≤-1 (x ∷ xs) with Byte-to-Bit8 x
    ... | (b0 ∷ b1 ∷ b2 ∷ b3 ∷ b4 ∷ b5 ∷ b6 ∷ 𝟘 ∷ [])
          = real
            ⟨ C.fromℕ (proj₁ (unbits {7} (b0 ∷ b1 ∷ b2 ∷ b3 ∷ b4 ∷ b5 ∷ b6 ∷ [])))
            , ≤-refl
            , xs ⟩
    ... | _ = cake

    decode-≤-2 : {m : ℕ} → Vec Byte (2 + m) → Maybe (Parse C.Char (2 + m))
    decode-≤-2 {m} x@(x₁ ∷ x₂ ∷ xs) with Byte-to-Bit8 x₁ ,′ Byte-to-Bit8 x₂
    ... | (b6 ∷ b7 ∷ b8 ∷ b9 ∷ b10 ∷ 𝟘  ∷ 𝟙 ∷ 𝟙 ∷ [])
        , (b0 ∷ b1 ∷ b2 ∷ b3 ∷ b4  ∷ b5 ∷ 𝟘 ∷ 𝟙 ∷ [])
          = real
            ⟨ C.fromℕ (proj₁ (unbits {11}
              ( b0 ∷ b1 ∷ b2  ∷ b3 ∷ b4 ∷ b5 ∷ b6 ∷ b7
              ∷ b8 ∷ b9 ∷ b10 ∷ []
              )))
            , m≤n+m (suc m) 1
            , xs ⟩
    ... | _ = cake

    decode-≤-3 : {m : ℕ} → Vec Byte (3 + m) → Maybe (Parse C.Char (3 + m))
    decode-≤-3 {m} x@(x₁ ∷ x₂ ∷ x₃ ∷ xs) with Byte-to-Bit8 x₁ ,′ Byte-to-Bit8 x₂ ,′ Byte-to-Bit8 x₃
    ... | (b12 ∷ b13 ∷ b14 ∷ b15 ∷ 𝟘   ∷ 𝟙   ∷ 𝟙 ∷ 𝟙 ∷ [])
        , (b6  ∷ b7  ∷ b8  ∷ b9  ∷ b10 ∷ b11 ∷ 𝟘 ∷ 𝟙 ∷ [])
        , (b0  ∷ b1  ∷ b2  ∷ b3  ∷ b4  ∷ b5  ∷ 𝟘 ∷ 𝟙 ∷ [])
          = real
            ⟨ C.fromℕ (proj₁ (unbits {16}
              ( b0 ∷ b1 ∷ b2  ∷ b3  ∷ b4  ∷ b5  ∷ b6  ∷ b7
              ∷ b8 ∷ b9 ∷ b10 ∷ b11 ∷ b12 ∷ b13 ∷ b14 ∷ b15 ∷ []
              )))
            , m≤n+m (suc m) 2
            , xs ⟩
    ... | _ = cake

    decode-≤-4 : {m : ℕ} → Vec Byte (4 + m) → Maybe (Parse C.Char (4 + m))
    decode-≤-4 {m} x@(x₁ ∷ x₂ ∷ x₃ ∷ x₄ ∷ xs) with Byte-to-Bit8 x₁ ,′ Byte-to-Bit8 x₂ ,′ Byte-to-Bit8 x₃ ,′ Byte-to-Bit8 x₄
    ... | (b18 ∷ b19 ∷ b20 ∷ 𝟘   ∷ 𝟙   ∷ 𝟙   ∷ 𝟙 ∷ 𝟙 ∷ [])
        , (b12 ∷ b13 ∷ b14 ∷ b15 ∷ b16 ∷ b17 ∷ 𝟘 ∷ 𝟙 ∷ [])
        , (b6  ∷ b7  ∷ b8  ∷ b9  ∷ b10 ∷ b11 ∷ 𝟘 ∷ 𝟙 ∷ [])
        , (b0  ∷ b1  ∷ b2  ∷ b3  ∷ b4  ∷ b5  ∷ 𝟘 ∷ 𝟙 ∷ [])
          = real
            ⟨ C.fromℕ (proj₁ (unbits {21}
              ( b0  ∷ b1  ∷ b2  ∷ b3  ∷ b4  ∷ b5  ∷ b6  ∷ b7
              ∷ b8  ∷ b9  ∷ b10 ∷ b11 ∷ b12 ∷ b13 ∷ b14 ∷ b15
              ∷ b16 ∷ b17 ∷ b18 ∷ b19 ∷ b20 ∷ []
              )))
            , m≤n+m (suc m) 3
            , xs ⟩
    ... | _ = cake

    parse-char : [ Parser C.Char ]
    parse-char {n} .parse [] = cake
    parse-char {n} .parse xs@(_ ∷ _) = case decode-≤-1 xs of λ where
      (real x) → real x
      cake → case xs of λ where
        (_ ∷ []) → cake
        (_ ∷ _ ∷ _) → case decode-≤-2 xs of λ where
          (real x) → real x
          cake → case xs of λ where
            (_ ∷ _ ∷ []) → cake
            (_ ∷ _ ∷ _ ∷ _) → case decode-≤-3 xs of λ where
              (real x) → real x
              cake → case xs of λ where
                (_ ∷ _ ∷ _ ∷ []) → cake
                (_ ∷ _ ∷ _ ∷ _ ∷ _) → decode-≤-4 xs

  parse-string : [ Parser (List⁺ C.Char) ]
  parse-string = many-fix parse-char

  decode-string : List Byte → Maybe String
  decode-string L.[] = real ""
  decode-string xs@(_ L.∷ _) with parse-string .parse (V.fromList xs)
  ... | real ⟨ r , _ , V.[] ⟩ = real (S.fromList (list⁺-to-list r))
  ... | _ = cake

open decode using (decode-string; parse-string; parse-char) public

